import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from time import sleep
from PIL import Image
from PIL import ImageDraw
from PIL import ImageFont
from PIL import ImageOps
from copreus.baseclasses.aepaper import EPaperConstants
from copreus.drivers.epapersimple import EPaperSimple, EPaperMQTTMessageConverter
from pelops.myconfigtools import read_config, validate_config
from pelops.mylogger import create_logger
from pelops.mymqttclient import MyMQTTClient
import threading
import time
import datetime


class MessageQueueStats:
    max_count = None
    min_count = None
    akt_count = None
    count_list = None
    message_counter = 0
    _event = None
    TIMEOUT = 10
    RAISE_EXCEPTIONS = False

    def __init__(self, mqtt_client, topic, event):
        self.count_list = []
        mqtt_client.subscribe(topic, self._message_queue_size_handler)
        self._event = event

    def to_string(self):
        str = "count: {}; min: {}, max: {}, akt: {}, list: {}."\
            .format(self.message_counter, self.max_count, self.min_count, self.akt_count, self.count_list)
        return str

    def _message_queue_size_handler(self, value):
        value = int(value)
        self.message_counter += 1
        self.akt_count = value
        if self.max_count is None:
            self.min_count = value
            self.max_count = value
        self.min_count = min(value, self.min_count)
        self.max_count = max(value, self.max_count)
        self.count_list.append(value)
        self._event.set()

    def wait_for_message(self, expected_value):
        print("wait_for_message. '{}'".format(expected_value))
        self._event.wait(timeout=self.TIMEOUT)
        if not self._event.is_set():
            print("RuntimeError - wait_for_message({}) timeout occurred.".format(expected_value))
            if self.RAISE_EXCEPTIONS:
                raise RuntimeError("wait_for_message({}) timeout occurred.".format(expected_value))
        if self.akt_count != expected_value:
            print("ValueError - expected: {} != received: {}".format(expected_value, self.akt_count))
            if self.RAISE_EXCEPTIONS:
                raise ValueError("expected: {} != received: {}".format(expected_value, self.akt_count))
        self._event.clear()

    def wait_for_value(self, expected_value, max_counter=10):
        print("wait_for_value. '{}'".format(expected_value))
        self._event.wait(timeout=self.TIMEOUT)
        if not self._event.is_set():
            print("RuntimeError - wait_for_value({}) timeout occurred.".format(expected_value))
            if self.RAISE_EXCEPTIONS:
                raise RuntimeError("wait_for_value({}) timeout occurred.".format(expected_value))
        counter = 0
        while self.akt_count != expected_value:
            counter += 1
            self._event.clear()
            self._event.wait(timeout=self.TIMEOUT)
            if not self._event.is_set():
                print("current count list: {}".format(self.count_list))
                print("RuntimeError - wait_for_value({}) timeout occurred.".format(expected_value))
                if self.RAISE_EXCEPTIONS:
                    raise RuntimeError("wait_for_value({}) timeout occurred.".format(expected_value))
            if counter > max_counter:
                print("current count list: {}".format(self.count_list))
                print("RuntimeError - wait_for_value({}) max counter violation occurred ({}>{})."
                      .format(expected_value, counter, max_counter))
                if self.RAISE_EXCEPTIONS:
                    raise RuntimeError("wait_for_value({}) max counter violation occurred ({}>{})."
                                       .format(expected_value, counter, max_counter))
        self._event.clear()


filename = os.path.abspath(os.path.join(os.path.dirname(__file__), '..')) + "/tests_unit/config_epapersimple.yaml"
config = read_config(filename)
logger = create_logger(config["logger"], "TestEPaperSimple")
logger.info("start =====================================================================================")
mqtt_client = MyMQTTClient(config["mqtt"], logger)
mqtt_client.connect()


# sudo apt install fonts-freefont-ttf
font_file = "/usr/share/fonts/truetype/freefont/FreeMonoBold.ttf"

model = str(config["driver"]["model"])
display_width = EPaperConstants.model[model]["height"]
display_height = EPaperConstants.model[model]["width"]
topic_sub_image = config["driver"]["topics-sub"]["image"]

stats_event = threading.Event()
stats_event.clear()
stats = MessageQueueStats(mqtt_client, config["driver"]["topics-pub"]["message_queue_size"], stats_event)

logger.info("--- wiping screen at start-up -------------------------------------------------------")
print("--- wiping screen at start-up ---")
config["driver"]["wipe-screen"]["every-nth-day"] = 0
config["driver"]["wipe-screen"]["time"] = "00:00"
config["driver"]["wipe-screen"]["at-start-up"] = True
eps = EPaperSimple(config)
eps_thread = threading.Thread(target=eps.run)
start = time.time()
eps_thread.start()
eps._is_started.wait()
stats.wait_for_message(1)
stats.wait_for_message(0)
eps.stop()
eps.is_stopped.wait()
eps_thread.join()
stop = time.time()
akt_diff = stop-start
expected_min_diff = eps._update_time * 6  # six color changes
print("... possibly a success: {}. measured {} seconds. expected minimum {} seconds."
      .format((expected_min_diff<=akt_diff), akt_diff, expected_min_diff))
time.sleep(1)

logger.info("--- sending two images -------------------------------------------------------")
print("--- sending two images ---")
config["driver"]["wipe-screen"]["every-nth-day"] = 0
config["driver"]["wipe-screen"]["time"] = "00:00"
config["driver"]["wipe-screen"]["at-start-up"] = False
eps = EPaperSimple(config)
eps_thread = threading.Thread(target=eps.run)
eps_thread.start()
eps._is_started.wait()

try:
    font_small = ImageFont.truetype(font_file, 12)
    font_large = ImageFont.truetype(font_file, 64)
except OSError:
    print("OS Error - can't open font file '{}'.".format(font_file))
    raise

regular_image = Image.new('L', (display_width, display_height), 255)  # 255: clear the frame
draw = ImageDraw.Draw(regular_image)
draw.rectangle((0, 10, display_width, 32), fill=0)
draw.text((30, 16), 'e-Paper Demo', font=font_small, fill=255)

print("... sending regular image")
msg = EPaperMQTTMessageConverter.to_full_image(regular_image)
mqtt_client.publish(topic_sub_image, msg)
stats.wait_for_message(1)
stats.wait_for_message(0)

sleep(1)

print("... sending inverted image")
inverted_image = ImageOps.invert(regular_image)
msg = EPaperMQTTMessageConverter.to_full_image(inverted_image)
mqtt_client.publish(topic_sub_image, msg)
stats.wait_for_message(1)
stats.wait_for_message(0)

eps.stop()
eps.is_stopped.wait()
eps_thread.join()

logger.info("--- test message-queue -------------------------------------------------------")
print("--- test message-queue (sending 4 regular and 1 inverted image - 3 regular images should be discarded) ---")
config["driver"]["wipe-screen"]["every-nth-day"] = 0
config["driver"]["wipe-screen"]["time"] = "00:00"
config["driver"]["wipe-screen"]["at-start-up"] = False

msg_regular = EPaperMQTTMessageConverter.to_full_image(regular_image)
msg_inverted = EPaperMQTTMessageConverter.to_full_image(inverted_image)

eps = EPaperSimple(config)
eps_thread = threading.Thread(target=eps.run)
eps_thread.start()
eps._is_started.wait()
start = time.time()

mqtt_client.publish(topic_sub_image, msg_regular)
mqtt_client.publish(topic_sub_image, msg_regular)
mqtt_client.publish(topic_sub_image, msg_regular)
mqtt_client.publish(topic_sub_image, msg_regular)
mqtt_client.publish(topic_sub_image, msg_regular)
mqtt_client.publish(topic_sub_image, msg_regular)
mqtt_client.publish(topic_sub_image, msg_regular)
mqtt_client.publish(topic_sub_image, msg_regular)
mqtt_client.publish(topic_sub_image, msg_inverted)

stats.wait_for_value(0)
diff = time.time() - start
print("sended 9 images and displayed 2 in {} seconds. current stats: {}".format(diff, stats.to_string()))
if stats.count_list != [1, 0, 1, 0, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0]:
    raise RuntimeError("Not expected sequence of message queue sizes. received {}; "
                       "expected [1, 0, 1, 0, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0].".format(stats.count_list))

eps.stop()
eps.is_stopped.wait()
eps_thread.join()


logger.info("--- wipe screen timer -------------------------------------------------------")
print("--- wipe screen timer (restores to regular image) ---")
# set time for wipe screen to next full minute that is at least 10 seconds away (but not longer than 60 seconds)
now = time.time()
next_full_minute = 60 - (now % 60)
if next_full_minute < 20:
    next_full_minute += 60
now = datetime.datetime.now()
target = now + datetime.timedelta(seconds=next_full_minute)
target_time = "{:02d}:{:02d}".format(target.hour, target.minute)
print("... wipe screen scheduled to: {} / in {} seconds.".format(target_time, next_full_minute))
config["driver"]["wipe-screen"]["every-nth-day"] = 1
config["driver"]["wipe-screen"]["time"] = target_time
config["driver"]["wipe-screen"]["at-start-up"] = False

eps = EPaperSimple(config)
eps_thread = threading.Thread(target=eps.run)
eps_thread.start()
eps._is_started.wait()
mqtt_client.publish(topic_sub_image, msg_regular)

print("... waiting for approximately {:0.2f} seconds.".format(next_full_minute))

temp = stats.TIMEOUT
stats.TIMEOUT=next_full_minute+30
stats.wait_for_message(1)
stats.TIMEOUT = temp
stats.wait_for_value(0)

eps.stop()
eps.is_stopped.wait()
eps_thread.join()

print("... finished wipe screen timer")

print("final stats: ".format(stats.count_list))
if stats.max_count != 1 and stats.min_count != 0:
    raise RuntimeError("states are not as expected: max {} vs. max target 1 and min {} vs. min target 0."
                       .format(stats.max_count, stats.min_count))


logger.info("--- finished testing -------------------------------------------------------")
print("--- finished testing ---")

print("cleanup")
sleep(0.5)
mqtt_client.disconnect()
sleep(0.5)
print("done")
logger.info("stop ======================================================================================")


import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from time import sleep, time, strftime
import _thread
from PIL import Image
from PIL import ImageDraw
from PIL import ImageFont
from PIL import ImageOps
from copreus.baseclasses.aepaper import EPaperConstants
from copreus.drivers.epaperdirect import EPaperDirect, EPaperMQTTMessageConverter
from pelops.myconfigtools import read_config, validate_config
from pelops.mylogger import create_logger
from pelops.mymqttclient import MyMQTTClient
import threading
import RPi.GPIO as GPIO


class MessageQueueStats:
    max_count = None
    min_count = None
    akt_count = None
    count_list = None
    message_counter = 0
    _event = None
    TIMEOUT = 5
    RAISE_EXCEPTIONS = False

    def __init__(self, mqtt_client, topic, event):
        self.count_list = []
        mqtt_client.subscribe(topic, self._message_queue_size_handler)
        self._event = event

    def _message_queue_size_handler(self, value):
        value = int(value)
        self.message_counter += 1
        self.akt_count = value
        if self.max_count is None:
            self.min_count = value
            self.max_count = value
        self.min_count = min(value, self.min_count)
        self.max_count = max(value, self.max_count)
        self.count_list.append(value)
        self._event.set()

    def wait_for_message(self, expected_value):
        self._event.wait(timeout=self.TIMEOUT)
        if not self._event.is_set():
            print("RuntimeError - wait_for_message({}) timeout occurred.".format(expected_value))
            if self.RAISE_EXCEPTIONS:
                raise RuntimeError("wait_for_message({}) timeout occurred.".format(expected_value))
        if self.akt_count != expected_value:
            print("ValueError - expected: {} != received: {}".format(expected_value, self.akt_count))
            if self.RAISE_EXCEPTIONS:
                raise ValueError("expected: {} != received: {}".format(expected_value, self.akt_count))
        self._event.clear()

    def wait_for_value(self, expected_value, max_counter=10):
        self._event.wait(timeout=self.TIMEOUT)
        if not self._event.is_set():
            print("RuntimeError - wait_for_value({}) timeout occurred.".format(expected_value))
            if self.RAISE_EXCEPTIONS:
                raise RuntimeError("wait_for_value({}) timeout occurred.".format(expected_value))
        counter = 0
        while self.akt_count != expected_value:
            counter += 1
            self._event.clear()
            self._event.wait(timeout=self.TIMEOUT)
            if not self._event.is_set():
                print("current count list: {}".format(self.count_list))
                print("RuntimeError - wait_for_value({}) timeout occurred.".format(expected_value))
                if self.RAISE_EXCEPTIONS:
                    raise RuntimeError("wait_for_value({}) timeout occurred.".format(expected_value))
            if counter > max_counter:
                print("current count list: {}".format(self.count_list))
                print("RuntimeError - wait_for_value({}) max counter violation occurred ({}>{})."
                      .format(expected_value, counter, max_counter))
                if self.RAISE_EXCEPTIONS:
                    raise RuntimeError("wait_for_value({}) max counter violation occurred ({}>{})."
                                       .format(expected_value, counter, max_counter))
        self._event.clear()


filename = os.path.abspath(os.path.join(os.path.dirname(__file__), '..')) + "/tests_unit/config_epaperdirect.yaml"
config = read_config(filename)
logger = create_logger(config["logger"], "TestEPaperDirect")
logger.info("start")
mqtt_client = MyMQTTClient(config["mqtt"], logger)
mqtt_client.connect()

# sudo apt install fonts-freefont-ttf
font_file = "/usr/share/fonts/truetype/freefont/FreeMonoBold.ttf"

model = str(config["driver"]["model"])
display_width = EPaperConstants.model[model]["height"]
display_height = EPaperConstants.model[model]["width"]
topic_sub_full_image = config["driver"]["topics-sub"]["full_image"]
topic_sub_partial_image = config["driver"]["topics-sub"]["partial_image"]
topic_sub_switch_frame = config["driver"]["topics-sub"]["switch_frame"]
command_switch_frame = config["driver"]["mqtt-translations"]["switch_frame"]

stats_event = threading.Event()
stats_event.clear()
stats = MessageQueueStats(mqtt_client, config["driver"]["topics-pub"]["message_queue_size"], stats_event)

_ep_thread = _thread.start_new_thread(EPaperDirect.standalone, (('-c', filename),))

try:
    font_small = ImageFont.truetype(font_file, 12)
    font_large = ImageFont.truetype(font_file, 64)
except OSError:
    print("OS Error - can't open font file '{}'.".format(font_file))
    raise

sleep(1)
stats.wait_for_message(0)

base_image = Image.new('L', (display_width, display_height), 255)  # 255: clear the frame
draw = ImageDraw.Draw(base_image)
draw.rectangle((0, 10, display_width, 32), fill=0)
draw.text((30, 16), 'e-Paper Demo', font=font_small, fill=255)
inverted_image = ImageOps.invert(base_image)

print("--- sending full image once ---")
msg = EPaperMQTTMessageConverter.to_full_image(base_image)
mqtt_client.publish(topic_sub_full_image, msg)
stats.wait_for_message(1)
sleep(5)
stats.wait_for_message(0)

print("--- sending inverted full image once ---")
msg = EPaperMQTTMessageConverter.to_full_image(inverted_image)
mqtt_client.publish(topic_sub_full_image, msg)
stats.wait_for_message(1)
sleep(5)
stats.wait_for_message(0)

print("--- switching display frames 4 times ---")
for i in range(4):
    print(i, "SWITCH")
    mqtt_client.publish(topic_sub_switch_frame, "SWITCH")
    stats.wait_for_message(1)
    sleep(5)
    stats.wait_for_message(0)

print("--- test message queue ---")
print("... sending 4 images")
msg = EPaperMQTTMessageConverter.to_full_image(base_image)
mqtt_client.publish(topic_sub_full_image, msg)
mqtt_client.publish(topic_sub_full_image, msg)
mqtt_client.publish(topic_sub_full_image, msg)
mqtt_client.publish(topic_sub_full_image, msg)
print("expecting queues size 4")
stats.wait_for_value(4, 5)
if stats.max_count != 4:
    raise ValueError("stats max_counter {} not as expected {}.".format(stats.max_count, 4))
print("expecting queues size 3")
stats.wait_for_message(3)
print("expecting queues size 2")
stats.wait_for_message(2)
print("expecting queues size 1")
stats.wait_for_message(1)
print("expecting queues size 0")
stats.wait_for_message(0)
print("final stats: {}".format(stats.count_list))
sleep(5)

print("--- sending full image twice ---")
print("... 1")
msg = EPaperMQTTMessageConverter.to_full_image(base_image)
mqtt_client.publish(topic_sub_full_image, msg)
stats.wait_for_message(1)
stats.wait_for_message(0)
print("... 2")
msg = EPaperMQTTMessageConverter.to_full_image(base_image)
mqtt_client.publish(topic_sub_full_image, msg)
stats.wait_for_message(1)
stats.wait_for_message(0)

sleep(5)

print("--- testing partial image ---")
time_large = Image.new('1', (205, 80), 255)  # 255: clear the frame
draw_large = ImageDraw.Draw(time_large)
time_small = Image.new('1', (56, 16), 0)
draw_small = ImageDraw.Draw(time_small)

large_width, large_height = time_large.size
small_width, small_height = time_small.size

minutes = 3
print("... displaying time for the next {} minutes".format(minutes))
for i in range(minutes):
    t = strftime('%H:%M')

    draw_large.rectangle((0, 0, large_width, large_height), fill=255)
    draw_large.text((0, 0), t, font=font_large, fill=0)

    draw_small.rectangle((0, 0, small_width, small_height), fill=0)
    draw_small.text((0,0), t, font=font_small, fill=255)

    list = [
        {
            "x": int((display_width-large_width)/2),
            "y": 40,
            "image": time_large
        },
        {
            "x": display_width-small_width-16,
            "y": 16,
            "image": time_small
        }
    ]
    msg = EPaperMQTTMessageConverter.to_partial_images(list)
    print("updating time to '{}'".format(t))
    mqtt_client.publish(topic_sub_partial_image, msg)
    stats.wait_for_message(1)

    ct = time()
    current_second = int(ct) * 1000
    current_ms = int(ct * 1000.0)

    step_target = 60
    step = (step_target - ct % step_target)
    step_safety = 1

    diff = step + ((current_ms - current_second) / 1000.0) # wait this for the next ms ...

    print("sleeping for {} seconds.".format(diff+step_safety))
    sleep(diff + step_safety)
    stats.wait_for_message(0)

print("--- stopped testing partial image ---")

print("--- erase ---")
print("... sending black and white images. takes approx. 52 seconds.")

white_image = Image.new('1', (EPaperConstants.model[model]["height"], EPaperConstants.model[model]["width"]), 255)  # 255: clear the frame
black_image = Image.new('1', (EPaperConstants.model[model]["height"], EPaperConstants.model[model]["width"]), 0)  # 0: clear the frame
white_draw = ImageDraw.Draw(white_image)
black_draw = ImageDraw.Draw(black_image)
msg_white = EPaperMQTTMessageConverter.to_full_image(white_image)
msg_black = EPaperMQTTMessageConverter.to_full_image(black_image)

for i in range(4):
    print("erase loop {}.".format(i))
    print("sending black image a")
    mqtt_client.publish(topic_sub_full_image, msg_black)
    sleep(1)
    print("sending black image b")
    mqtt_client.publish(topic_sub_full_image, msg_black)
    sleep(5)
    print("sending white image a")
    mqtt_client.publish(topic_sub_full_image, msg_white)
    sleep(1)
    print("sending white image b")
    mqtt_client.publish(topic_sub_full_image, msg_white)
    sleep(5)

stats.wait_for_value(0)
print("final stats: ".format(stats.count_list))
print("--- finished testing ---")

print("cleanup")
sleep(0.5)
mqtt_client.disconnect()
GPIO.cleanup()
sleep(0.5)
print("done")



import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import _thread
from time import sleep

from copreus.drivermanager.drivermanager import DriverManager
from tests_manual.tools import send_to_topic, get_topics_sub, get_resolution, get_maxvalue, resister_to_pubs, \
    get_mqtt_translations


filename = '../tests_unit/config_adcdac.yaml'
dac_pos = 1
adc_pos = 0


def loop(topics):
    adc_topic = get_topics_sub(filename, adc_pos)["readnow"]
    adc_command = get_mqtt_translations(filename, adc_pos)["readnow"]
    while 1:
        topic = topics["raw"]
        span = get_resolution(filename, dac_pos)
        steps = 10
        step_size = float(span) / float(steps)
        for i in range(0, steps+1):
            value = int(step_size * i)
            send_to_topic(filename, topic, value)
            sleep(1)
            send_to_topic(filename, adc_topic, adc_command)
            sleep(1)

        topic = topics["volt"]
        span = get_maxvalue(filename, dac_pos)
        steps = 10
        step_size = float(span) / float(steps)
        for i in range(0, steps+1):
            value = (step_size * i)
            send_to_topic(filename, topic, value)
            sleep(1)
            send_to_topic(filename, adc_topic, adc_command)
            sleep(1)


subs = get_topics_sub(filename, dac_pos)
resister_to_pubs(filename, adc_pos)
_loop_thread = _thread.start_new_thread(loop, (subs,))

# -------------- above are preparations - below is starting of the driver manager

DriverManager.standalone(('-c', filename))

#_config, _verbose = DriverManager._args_to_config(('-c', filename))
#dm = DriverManager(_config, _verbose)
#dm._start()
#try:
#    while not dm._is_stopped.wait(0.1):  # timeout is necessary for CTRL+C
#        pass
#except KeyboardInterrupt:
#    pass
#dm._stop()


import copreus.schema.adriver as adriver
import copreus.schema.apolling as apolling
import copreus.schema.aspi as aspi
import copreus.schema.calibratedvalue as calibratedvalue


def _add_schema_part(schema, schema_part):
    schema["driver"]["required"].extend(schema_part["required"])
    schema["driver"]["properties"].update(schema_part["properties"])


def get_schema():
    driver_specific_properties = {
        "pin": {
            "description": "gpio pin",
            "type": "integer"
        },
        "physical-closed": {
            "description": " high/low - mapping between logcial states (closed/open) and physical "
                           "output parameters (low/high)",
            "type": "string",
            "enum": ["low", "high"]
        },
        "initially-closed": {
            "description": "True/False - defines if the output is opened or closed after start of driver",
            "type": "boolean"
        }
    }

    topics_pub = {
    }
    topics_sub = {
        "closed": " mqtt-translations.closed-true and mqtt-translations.closed-false"
    }
    mqtt_translations = {
        "closed-true": "value for closed output",
        "closed-false": "value for open output",
    }

    schema = adriver.get_schema("output", driver_specific_properties, topics_pub, topics_sub, mqtt_translations)

    return schema



import unittest
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from pelops.myconfigtools import read_config, validate_config
import pelops.schema.abstractmicroservice


class TestValidateConfig(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.prefix = ""
        if not os.getcwd().endswith("tests_unit"):
            cls.prefix = "tests_unit/"

    def test_validate_adcdac(self):
        import copreus.schema.drivermanager as schema
        config = read_config(self.prefix + "config_adcdac.yaml")
        validation_result = validate_config(config, pelops.schema.abstractmicroservice.get_schema(schema.get_schema()))
        self.assertIsNone(validation_result)

    def test_validate_adc(self):
        import copreus.schema.adc as schema
        config = read_config(self.prefix + "config_adc.yaml")
        validation_result = validate_config(config, pelops.schema.abstractmicroservice.get_schema(schema.get_schema()))
        self.assertIsNone(validation_result)

    def test_validate_dac(self):
        import copreus.schema.dac as schema
        config = read_config(self.prefix + "config_dac.yaml")
        validation_result = validate_config(config, pelops.schema.abstractmicroservice.get_schema(schema.get_schema()))
        self.assertIsNone(validation_result)

    def test_validate_bme_280(self):
        import copreus.schema.bme_280 as schema
        config = read_config(self.prefix + "config_bme_280.yaml")
        validation_result = validate_config(config, pelops.schema.abstractmicroservice.get_schema(schema.get_schema()))
        self.assertIsNone(validation_result)

    def test_validate_dht(self):
        import copreus.schema.dht as schema
        config = read_config(self.prefix + "config_dht.yaml")
        validation_result = validate_config(config, pelops.schema.abstractmicroservice.get_schema(schema.get_schema()))
        self.assertIsNone(validation_result)

    def test_validate_epaperdirect(self):
        import copreus.schema.epaperdirect as schema
        config = read_config(self.prefix + "config_epaperdirect.yaml")
        validation_result = validate_config(config, pelops.schema.abstractmicroservice.get_schema(schema.get_schema()))
        self.assertIsNone(validation_result)

    def test_validate_epapersimple(self):
        import copreus.schema.epapersimple as schema
        config = read_config(self.prefix + "config_epapersimple.yaml")
        validation_result = validate_config(config, pelops.schema.abstractmicroservice.get_schema(schema.get_schema()))
        self.assertIsNone(validation_result)

    def test_validate_input(self):
        import copreus.schema.input as schema
        config = read_config(self.prefix + "config_input.yaml")
        validation_result = validate_config(config, pelops.schema.abstractmicroservice.get_schema(schema.get_schema()))
        self.assertIsNone(validation_result)

    def test_validate_pollinginput(self):
        import copreus.schema.pollinginput as schema
        config = read_config(self.prefix + "config_pollinginput.yaml")
        validation_result = validate_config(config, pelops.schema.abstractmicroservice.get_schema(schema.get_schema()))
        self.assertIsNone(validation_result)

    def test_validate_output(self):
        import copreus.schema.output as schema
        config = read_config(self.prefix + "config_output.yaml")
        validation_result = validate_config(config, pelops.schema.abstractmicroservice.get_schema(schema.get_schema()))
        self.assertIsNone(validation_result)

    def test_validate_rotaryencoder(self):
        import copreus.schema.rotaryencoder as schema
        config = read_config(self.prefix + "config_rotaryencoder.yaml")
        validation_result = validate_config(config, pelops.schema.abstractmicroservice.get_schema(schema.get_schema()))
        self.assertIsNone(validation_result)

    def test_validate_rotaryencoder2(self):
        import copreus.schema.rotaryencoder2 as schema
        config = read_config(self.prefix + "config_rotaryencoder2.yaml")
        validation_result = validate_config(config, pelops.schema.abstractmicroservice.get_schema(schema.get_schema()))
        self.assertIsNone(validation_result)

    def test_validate_drivermanager(self):
        import copreus.schema.drivermanager as schema
        config = read_config(self.prefix + "config.yaml")
        validation_result = validate_config(config, pelops.schema.abstractmicroservice.get_schema(schema.get_schema()))
        self.assertIsNone(validation_result)


if __name__ == '__main__':
    unittest.main()



import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import _thread
from time import sleep

from copreus.drivermanager.drivermanager import DriverManager
from pelops.myconfigtools import read_config
from pelops.mymqttclient import MyMQTTClient
from pelops.mylogger import create_logger


def output_loop(topics, cmds):
    sleep(5)
    topic = topics["closed"]
    cmd_true = cmds["closed-true"]
    cmd_false = cmds["closed-false"]
    state = 0
    while 1:
        state = (state + 1) % 2
        if state:
            cmd = cmd_false
        else:
            cmd = cmd_true
        print(cmd)
        mqttclient.publish(topic, cmd)
        sleep(3)


def input_handler_pressed(message):
    print("button_pressed: {}".format(message))


def input_handler_state(message):
    print("button_state: {}".format(message))


def bme280_handler_temperature(message):
    print("bme280-temperature: {}".format(message))


def bme280_handler_humidity(message):
    print("bme280-humidity: {}".format(message))


def bme280_handler_pressure(message):
    print("bme280-pressure: {}".format(message))


def dht_handler_temperature(message):
    print("dht-temperature: {}".format(message))


def dht_handler_humidity(message):
    print("dht-humidity: {}".format(message))


filename = '../tests_unit/config_drivermanager.yaml'
config = read_config(filename)

name_to_listpos = {}
for pos in range(len(config["drivers"])):
    name_to_listpos[config["drivers"][pos]["name"]] = pos

logger = create_logger(config["logger"], __name__)
mqttclient = MyMQTTClient(config["mqtt"], logger, True)
mqttclient.connect()
mqttclient.subscribe(config["drivers"][name_to_listpos["input"]]["topics-pub"]["button_pressed"],
                     input_handler_pressed)
mqttclient.subscribe(config["drivers"][name_to_listpos["input"]]["topics-pub"]["button_state"],
                     input_handler_state)

mqttclient.subscribe(config["drivers"][name_to_listpos["bme280"]]["topics-pub"]["temperature"],
                     bme280_handler_temperature)
mqttclient.subscribe(config["drivers"][name_to_listpos["bme280"]]["topics-pub"]["humidity"],
                     bme280_handler_humidity)
mqttclient.subscribe(config["drivers"][name_to_listpos["bme280"]]["topics-pub"]["pressure"],
                     bme280_handler_pressure)

mqttclient.subscribe(config["drivers"][name_to_listpos["dht"]]["topics-pub"]["temperature"],
                     dht_handler_temperature)
mqttclient.subscribe(config["drivers"][name_to_listpos["dht"]]["topics-pub"]["humidity"],
                     dht_handler_humidity)

_output_loop_thread = _thread.start_new_thread(output_loop,
                                               (config["drivers"][name_to_listpos["output"]]["topics-sub"],
                                                config["drivers"][name_to_listpos["output"]]["mqtt-translations"],))

# -------------- above are preparations - below is starting of the driver manager

DriverManager.standalone(('-c', filename))




import copreus.drivers
import importlib
import pelops.mylogger as mylogger


class DriverFactory:
    factory_logger = None
    mqtt_client = None
    logger = None
    spi_lock = None
    i2c_lock = None
    #driver = None

    def __init__(self, mqtt_client=None, logger=None, spi_lock=None, i2c_lock=None):
        self.mqtt_client = mqtt_client
        self.logger = logger
        self.spi_lock = spi_lock
        self.i2c_lock = i2c_lock
        self.factory_logger = mylogger.get_child(logger, self.__class__.__name__)

    def create_drivers(self, config):
        """Static drivers factory - takes a list of driver configs and creates them."""

        self.factory_logger.info("creating drivers")
        self.factory_logger.debug("driver configs: ".format(config["drivers"]))
        drivers = {}
        for driver_config in config["drivers"]:
            if not driver_config["active"]:
                continue
            config["driver"] = driver_config
            driver = self.create_driver(config)
            driver_id = self.get_unique_name(drivers, driver.get_name())
            drivers[driver_id] = driver
            self.factory_logger.info(" - added driver '[{}]: {}.{}'".format(driver_id, driver._type, driver._name))
        self.factory_logger.debug("drivers: {}".format(drivers))
        return drivers

    @staticmethod
    def get_unique_name(drivers, name):
        prefix = name.replace(" ", "").lower()
        suffix = ""
        counter = 0
        name = prefix+suffix
        while name in drivers:
            suffix = "_{}".format(counter)
            counter += 1
            name = prefix + suffix
        return name

    def create_driver(self, config):
        """Static driver factory - takes a driver entry from json/yaml config and instantiates the corresponding
        Class. Classes that are specializations of ASPI are provided with the spi_lock (if one is provided to this
        factory).

        New implemented driver must be added manually."""
        type_name = config["driver"]["type"].upper()

        # it is on purpose that not class names are used to be compared with type_name (as will be done within the
        # constructor of the base class ADriver). This approach allows for late binding - thus, a class is imported
        # if and only if it is needed which results in less dependencies that must be fullfilled although they might
        # not be needed.

        drivers = copreus.drivers.get_drivers()

        try:
            driver = drivers[type_name]
            mod = importlib.import_module(driver["module"])
            klass = getattr(mod, driver["name"])
            if "ASPI" in driver["bases"]:
                result = klass(config, self.mqtt_client, self.logger, spi_lock=self.spi_lock, no_gui=True,
                               manage_monitoring_agent=False)
            elif "AI2C" in driver["bases"]:
                result = klass(config, self.mqtt_client, self.logger, i2c_lock=self.i2c_lock, no_gui=True,
                               manage_monitoring_agent=False)
            else:
                result = klass(config, self.mqtt_client, self.logger, no_gui=True,
                               manage_monitoring_agent=False)
        except:
            logger = mylogger.get_child(self.logger, __name__)
            logger.error("unknown type name '{}'.".format(type_name))
            raise ValueError("unknown type name '{}'.".format(type_name))

        return result

    def reload_driver(self, drivers, driver_id):
        """destroy the driver and creates a new one from scratch"""
        self.factory_logger.info("restarting driver '{}' - start".format(driver_id))
        driver = drivers[driver_id]
        self.factory_logger.info("... {}".format(driver.get_short_info()))
        driver.stop()
        config = {}
        config["driver"] = driver.get_config()
        new_driver = self.create_driver(config)
        drivers[driver_id] = new_driver
        new_driver.start()
        self.factory_logger.info("restarting driver {} - done")

    @staticmethod
    def old_create(config, mqtt_client, logger, spi_lock=None, i2c_lock=None):
        type_name = config["type"].upper()
        if type_name == "ADC":
            from copreus.drivers.adc import ADC
            result = ADC(config, mqtt_client, logger, spi_lock, no_gui=True, manage_monitoring_agent=False)
        elif type_name == "DAC":
            from copreus.drivers.dac import DAC
            result = DAC(config, mqtt_client, logger, spi_lock, no_gui=True, manage_monitoring_agent=False)
        elif type_name == "BME_280":
            from copreus.drivers.bme_280 import BME_280
            result = BME_280(config, mqtt_client, logger, i2c_lock, no_gui=True, manage_monitoring_agent=False)
        elif type_name == "DHT":
            from copreus.drivers.dht import DHT
            result = DHT(config, mqtt_client, logger, no_gui=True, manage_monitoring_agent=False)
        elif type_name == "EPAPERDIRECT":
            from copreus.drivers.epaperdirect import EPaperDirect
            result = EPaperDirect(config, mqtt_client, logger, spi_lock, no_gui=True, manage_monitoring_agent=False)
        elif type_name == "EPAPERSIMPLE":
            from copreus.drivers.epapersimple import EPaperSimple
            result = EPaperSimple(config, mqtt_client, logger, spi_lock, no_gui=True, manage_monitoring_agent=False)
        elif type_name == "INPUT":
            from copreus.drivers.input import Input
            result = Input(config, mqtt_client, logger, no_gui=True, manage_monitoring_agent=False)
        elif type_name == "POLLINGINPUT":
            from copreus.drivers.pollinginput import PollingInput
            result = PollingInput(config, mqtt_client, logger, no_gui=True, manage_monitoring_agent=False)
        elif type_name == "OUTPUT":
            from copreus.drivers.output import Output
            result = Output(config, mqtt_client, logger, no_gui=True, manage_monitoring_agent=False)
        elif type_name == "ROTARYENCODER":
            from copreus.drivers.rotaryencoder import RotaryEncoder
            result = RotaryEncoder(config, mqtt_client, logger, no_gui=True, manage_monitoring_agent=False)
        elif type_name == "ROTARYENCODER2":
            from copreus.drivers.rotaryencoder2 import RotaryEncoder2
            result = RotaryEncoder2(config, mqtt_client, logger, no_gui=True, manage_monitoring_agent=False)
        else:
            logger = mylogger.get_child(logger, __name__)
            logger.error("unknown type name '{}'.".format(type_name))
            raise ValueError("unknown type name '{}'.".format(type_name))

        return result





import RPi.GPIO as GPIO
import bme280 as bme280_driver  # https://pypi.python.org/pypi/RPi.bme280
from smbus2 import SMBus  # https://pypi.python.org/pypi/smbus2/0.2.0
from copreus.baseclasses.adriver import ADriver
from copreus.baseclasses.apolling import APolling
from copreus.baseclasses.ai2c import AI2C
from copreus.baseclasses.calibratedvalue import CalibratedValue
from copreus.schema.bme_280 import get_schema
from copreus.baseclasses.aevents import AEvents


class BME_280(ADriver, APolling, AEvents, AI2C):
    # BME280 would result in a name conflict with the above imported driver ...
    """Driver for the BME280 sensor (temperature, humidity, and pressure) with i2c connectivity.

    The driver entry in the yaml file consists of:
      * ADriver entries
        * topics_pub: temperature, humidity, pressure
      * APolling entries
      * i2c entries:
        * port: i2c port
        * address: i2c address
      * CalibratedValue entries in a sub-block named 'calibration_temperature'
      * CalibratedValue entries in a sub-block named 'calibration_humidity'
      * CalibratedValue entries in a sub-block named 'calibration_pressure'

    Example:
    driver:
        poll-interval: 30
        type: bme_280
        port: 1
        address: 0x76
        topics-pub:
            temperature: /bme280/temperature/raw
            humidity: /bme280/humidity/raw
            pressure: /bme280/pressure/raw
        topics-sub:
            poll-now: /bme280/pollnow
        mqtt-translations:
            poll-now: True
        calibration-temperature:
            use-calibration: True
            # - [ref_value, raw_value]
        calibration-humidity:
            use-calibration: True
            # - [ref_value, raw_value]
        calibration-pressure:
            use-calibration: True
            # - [ref_value, raw_value]
        event-pin:  # trigger for poll_now (optional)
            pin: 21
            flank: falling  # [falling, rising, both]
            topics-pub:  # optional
                button_pressed: /test/button/pressed
                button_state:   /test/button/state
            mqtt-translations:  # optional
                button_pressed: PRESSED
                button_state-open: OPEN
                button_state-closed: CLOSED

    Note: this class is named BME_280 due to a name conflict with the used Adafruit driver named bme280.
    """

    _port = -1  # i2c port
    _address = -1  # i2c address
    _bus = None  # SMBus instance
    _calibrated_t = None  # copreus.baseclasses.CalibratedValue for temperature
    _calibrated_h = None  # copreus.baseclasses.CalibratedValue for humidity
    _calibrated_p = None  # copreus.baseclasses.CalibratedValue for airpressure
    _event_pin = None  # input pin for poll_now trigger
    _event_flank_rising = None  # trigger poll_now on rising flank
    _event_flank_falling = None  # trigger poll_now on falling flank

    def __init__(self, config, mqtt_client=None, logger=None, i2c_lock=None, stdout_log_level=None, no_gui=None,
                 manage_monitoring_agent=True):
        ADriver.__init__(self, config, mqtt_client, logger, logger_name=self.__class__.__name__,
                         stdout_log_level=stdout_log_level, no_gui=no_gui,
                         manage_monitoring_agent=manage_monitoring_agent)
        APolling.__init__(self, self._config, self._mqtt_client, self._logger)
        AEvents.__init__(self, self._config, self._logger)
        AI2C.__init__(self, i2c_lock)

        self._port = self._config["port"]
        self._address = self._config["address"]

        self._calibrated_t = CalibratedValue(self._logger, self._config["calibration-temperature"], 1)
        self._calibrated_h = CalibratedValue(self._logger, self._config["calibration-humidity"], 1)
        self._calibrated_p = CalibratedValue(self._logger, self._config["calibration-pressure"], 1)

        self._ui_commands["poll"] = self._cmd_poll

        try:
            self._event_pin = self._config["event-pin"]["pin"]
            if self._config["event-pin"]["flank"] == "rising":
                self._event_flank_rising = True
                self._event_flank_falling = False
            elif self._config["event-pin"]["flank"] == "falling":
                self._event_flank_rising = False
                self._event_flank_falling = True
            elif self._config["event-pin"]["flank"] == "both":
                self._event_flank_rising = True
                self._event_flank_falling = True
            else:
                self._logger.error("BME_280.__init__ - unkown value for 'event-pin.flank' entry: '{}'.".
                                   format(self._config["event-pin"]["flank"]))
                raise ValueError("BME_280.__init__ - unkown value for 'event-pin.flank' entry: '{}'.".
                                 format(self._config["event-pin"]["flank"]))
            # assign both dicts for a variable before calling init methods - this ensures that either both are present
            # or both are missing
            topics_pub = self._config["event-pin"]["topics-pub"]
            mqtt_translations = self._config["event-pin"]["mqtt-translations"]
            self._add_topics_pub(topics_pub)
            self._add_mqtt_translations(mqtt_translations)
            self._add_event(self._event_pin, self._callback_pin)
        except KeyError:
            pass

    def _cmd_poll(self, args):
        """poll - polls the current values from the bme280 and outputs them (temperature, humidity, pressure): POLL"""
        t, h, p = self._get_values()
        print("[{}] temperature: {} °C, humidity: {} %, pressure: {} hPa".format(self._name, t, h, p))

    def _callback_pin(self, channel):
        """The event pins state has changed - should poll_now be executed"""
        state = GPIO.input(self._event_pin)
        self._logger.info("BME_280._callback_pin - received event. pin state: {}.".format(state))
        if not state:
            if self._event_flank_rising:
                self.poll_now()
            self._publish_value(self._topics_pub["button_pressed"], self._mqtt_translations["button_pressed"])
            self._publish_value(self._topics_pub["button_state"], self._mqtt_translations["button_state-closed"])
        else:
            if self._event_flank_falling:
                self.poll_now()
            self._publish_value(self._topics_pub["button_state"], self._mqtt_translations["button_state-open"])

    def _get_values(self):
        """polls the bme280 and returns the calibrated temperature, humidity, and pressure"""
        with self._i2c_lock:
            data = bme280_driver.sample(self._bus, self._address)

        t = self._calibrated_t.value(data.temperature)
        h = self._calibrated_t.value(data.humidity)
        p = self._calibrated_t.value(data.pressure)

        return t, h, p

    def _poll_device(self):
        """APolling._poll_device"""
        t, h, p = self._get_values()
        self._publish_value(self._topics_pub["temperature"], t)
        self._publish_value(self._topics_pub["humidity"], h)
        self._publish_value(self._topics_pub["pressure"], p)

    def _driver_start(self):
        """ADriver._driver_start"""
        self._bus = SMBus(self._port)
        bme280_driver.load_calibration_params(self._bus, self._address)
        self._start_polling()
        if self._event_pin:
            GPIO.setmode(GPIO.BCM)
            GPIO.setup(self._event_pin, GPIO.IN, pull_up_down=GPIO.PUD_UP)
            self._register_events()

    def _driver_stop(self):
        """ADriver._driver_stop"""
        self._stop_polling()
        self._bus.close()
        if self._event_pin:
            self._unregister_events()
            GPIO.cleanup(self._event_pin)

    @classmethod
    def _get_schema(cls):
        return get_schema()

    def _runtime_information(self):
        return {}

    def _config_information(self):
        return {}


def standalone():
    """Calls the static method BME_280.standalone()."""
    BME_280.standalone()


if __name__ == "__main__":
    BME_280.standalone()



import copreus.schema.aepaper as aepaper
import copreus.schema.adriver as adriver
import copreus.schema.apolling as apolling
import copreus.schema.aspi as aspi
import copreus.schema.calibratedvalue as calibratedvalue


def _add_schema_part(schema, schema_part):
    schema["driver"]["required"].extend(schema_part["required"])
    schema["driver"]["properties"].update(schema_part["properties"])


def get_schema():
    driver_specific_properties = aepaper.get_schema()

    driver_specific_properties["wipe-screen"] = {
        "description": "wipe (display black/white images) the screen regularly",
        "type": "object",
        "properties": {
            "every-nth-day": {
                "description": "run wipe screen every day, every second day. 0 for no scheduled wipe events",
                "type": "integer",
                "minimum": 0
            },
            "time": {
                "description": "HH:MM",
                "type": "string"
            },
            "at-start-up": {
                "description": "run wipe at startup",
                "type": "boolean"
            }
        },
        "required": ["every-nth-day", "time", "at-start-up"],
        "additionalItems": False
    }

    topics_pub = {
        "message_queue_size": "publishes the number of messages that wait to be processes.",
    }

    topics_sub = {
        "image": "a single image covering the whole display to be placed in the current buffer.",
    }

    mqtt_translations = {}

    schema = adriver.get_schema("epapersimple", driver_specific_properties, topics_pub, topics_sub, mqtt_translations)
    _add_schema_part(schema, aspi.get_schema_parts())

    return schema